
/*
  Credit to Luma for his original project the IOT Bonavita BV382510V kettle and
  his sketch that this is based on.
  ----------------------------  
  
  This sketch is intended to toggle a button in reponse to incoming MQTT messages.
  It is intended to be used with Home Assistant or any other home automation server
  that supports MQTT switch's.

  The state topic is only updated when the powerLED changes.
  I've been unable to get the ESP8266 to turn the kettle off, not sure why but
  don't really care as I can't really think of a reason why I would need to
  remotely turn it off when it turns off by itself once boiled.

  Birth and last will messages added for availability topics.

  HARDWARE NOTES:
    * The power LED is PWM contrlled so reading the Power Pin is done using
    the pulseIn() function for 10miliseconds.
    
  REQUIREMENTS: PubSubClient
  Large portions of the code here are based on the "Basic ESP8266 MQTT example" from
  PubSubClient by Nick O'Leary available at http://pubsubclient.knolleary.net/
*/
#include <FS.h>
// PubSubClient - Version: Latest 
#include <PubSubClient.h>
#include <ESP8266WiFi.h>
#include <PubSubClient.h>
#include <ESP8266mDNS.h>
#include <WiFiUdp.h>
#include <ArduinoOTA.h>
#include <WiFiManager.h>
#include <ESP8266WebServer.h>
#include <ESP8266WebServerSecure.h>
#include <DNSServer.h>
#include <ArduinoJson.h>  //version 5.13.2

const char* APNAME = ""; // The name of wifi manager Access Point
const char* APPASSWORD = "";  // The password for the wifi manager Access Point
char mqtt_server[40] = "";  // default MQTT server IP address
char mqtt_port[6] = "1883"; // default MQTT server port
char mqtt_user[20] = ""; // default MQTT Username if required. 
char mqtt_pass[20] = ""; // default MQTT password if required.
const char* mqtt_willTopic = "home/Kettle/LWT"; // LWT topic to publish to when ungracefully disconnected
const char* mqtt_willMessage = "Offline";
const char* mqtt_Announcemessage = "Online"; //Announce to LWT topic on connection

// This is the topics to be subscribed and published to
const char* mqtt_topic = "home/Kettle/switch"; // The command topic to recieve commands
const char* mqtt_state = "home/Kettle/state"; //The state topic to publish updates to
const char* mqtt_onMessage = "on"; // the on messaage
const char* mqtt_offMessage = "off"; // the off message

const int powerButton = D1; // the ESP pin to trigger on, go high on
const int powerMonitor = D0; // the the ESP pin that will monitor the kettle LED
const int powerButtonHoldTime = 6000; // How long the on button should be pressed for
const int powerButtonWaitTime = 1000; // How long to wait for the kettle to wake up and between button presses
bool shouldSaveConfig = false;
long lastMsg = 0;
int kettleState = 0;
int kettlePreviousState = 0;

WiFiClient espClient;
PubSubClient client(espClient);

void saveConfigCallback () {
  Serial.println("Should save config");
  shouldSaveConfig = true;
}

void setup_wifi() {
  delay(10);
  Serial.println();
  //clean FS, for testing
  //SPIFFS.format();

  //read configuration from FS json
  Serial.println("mounting FS...");

  if (SPIFFS.begin()) {
    Serial.println("mounted file system");
    if (SPIFFS.exists("/config.json")) {
      //file exists, reading and loading
      Serial.println("reading config file");
      File configFile = SPIFFS.open("/config.json", "r");
      if (configFile) {
        Serial.println("opened config file");
        size_t size = configFile.size();
        // Allocate a buffer to store contents of the file.
        std::unique_ptr<char[]> buf(new char[size]);

        configFile.readBytes(buf.get(), size);
        DynamicJsonBuffer jsonBuffer;
        JsonObject& json = jsonBuffer.parseObject(buf.get());
        json.printTo(Serial);
        if (json.success()) {
          Serial.println("\nparsed json");

          strcpy(mqtt_server, json["mqtt_server"]);
          strcpy(mqtt_port, json["mqtt_port"]);
          strcpy(mqtt_user, json["mqtt_user"]);
          strcpy(mqtt_pass, json["mqtt_pass"]);

        } else {
          Serial.println("failed to load json config");
        }
        configFile.close();
      }
    }
  } else {
    Serial.println("failed to mount FS");
  }
  //end read

  // The extra parameters to be configured (can be either global or just in the setup)
  // After connecting, parameter.getValue() will get you the configured value
  // id/name placeholder/prompt default length
  WiFiManagerParameter custom_mqtt_server("server", "mqtt server", mqtt_server, 40);
  WiFiManagerParameter custom_mqtt_port("port", "mqtt port", mqtt_port, 6);
  WiFiManagerParameter custom_mqtt_user("user", "mqtt user", mqtt_user, 20);
  WiFiManagerParameter custom_mqtt_pass("pass", "mqtt pass", mqtt_pass, 20);

  //WiFiManager
  //Local intialization. Once its business is done, there is no need to keep it around
  WiFiManager wifiManager;

  //set config save notify callback
  wifiManager.setSaveConfigCallback(saveConfigCallback);

  //set static ip comment for DHCP
  //wifiManager.setSTAStaticIPConfig(IPAddress(10,0,1,99), IPAddress(10,0,1,1), IPAddress(255,255,255,0));
  
  //add all your parameters here
  wifiManager.addParameter(&custom_mqtt_server);
  wifiManager.addParameter(&custom_mqtt_port);
  wifiManager.addParameter(&custom_mqtt_user);
  wifiManager.addParameter(&custom_mqtt_pass);

  //reset settings - for testing
  //wifiManager.resetSettings();

  //set minimu quality of signal so it ignores AP's under that quality
  //defaults to 8%
  //wifiManager.setMinimumSignalQuality();
  
  //sets timeout until configuration portal gets turned off
  //useful to make it all retry or go to sleep
  //in seconds
  wifiManager.setTimeout(600);

  //fetches ssid and pass and tries to connect
  //if it does not connect it starts an access point with the specified name
  //here  "AutoConnectAP"
  //and goes into a blocking loop awaiting configuration
  if (!wifiManager.autoConnect(APNAME, APPASSWORD)) {
    Serial.println("failed to connect and hit timeout");
    delay(3000);
    //reset and try again, or maybe put it to deep sleep
    ESP.reset();
    delay(5000);
  }

  //if you get here you have connected to the WiFi
  Serial.println("connected...yeey :)");

  //read updated parameters
  strcpy(mqtt_server, custom_mqtt_server.getValue());
  strcpy(mqtt_port, custom_mqtt_port.getValue());
  strcpy(mqtt_user, custom_mqtt_user.getValue());
  strcpy(mqtt_pass, custom_mqtt_pass.getValue());

  //save the custom parameters to FS
  if (shouldSaveConfig) {
    Serial.println("saving config");
    DynamicJsonBuffer jsonBuffer;
    JsonObject& json = jsonBuffer.createObject();
    json["mqtt_server"] = mqtt_server;
    json["mqtt_port"] = mqtt_port;
    json["mqtt_user"] = mqtt_user;
    json["mqtt_pass"] = mqtt_pass;

    File configFile = SPIFFS.open("/config.json", "w");
    if (!configFile) {
      Serial.println("failed to open config file for writing");
    }

    json.printTo(Serial);
    json.printTo(configFile);
    configFile.close();
    //end save
  }  
  
  Serial.println("");
  Serial.print("WiFi connected succesfully and assigned IP: ");
  Serial.println(WiFi.localIP());
}

/*  WiFi.begin(ssid, password);

  while (WiFi.status() != WL_CONNECTED) {
    delay(500);
    Serial.print(".");
  }
*/


void callback(char* topic, byte* payload, unsigned int length) {
  String inMessage;
  Serial.print("I've got a callback! ");
  for (int i = 0; i < length; i++) {
    inMessage += (char)payload[i];
  }
  Serial.print("Incoming message ");
  Serial.print(topic);
  Serial.print(": ");
  Serial.println(inMessage);

  if (inMessage == mqtt_onMessage) {
    Serial.println("Received MQTT power on message");
    
    if (pulseIn(powerMonitor, LOW, 10000) > 0) { //kettle is already boiling
      Serial.println("Kettle is already boiling.");
    }
    else { // kettle not boiling so try short press power to boil
      Serial.println("Toggling power on");
      digitalWrite(powerButton, HIGH);
      delay(powerButtonWaitTime);
      digitalWrite(powerButton, LOW);
      delay(powerButtonWaitTime);

      if (pulseIn(powerMonitor, LOW, 10000) > 0) {  // power LED is on, kettle is boiling
        Serial.println("Kettle is boiling.");
      }
      else { // kettle is in standby mode
        Serial.println("Kettle is in stanby mode, attempting to wake up");
        digitalWrite(powerButton, HIGH); // wake it back up
        delay(powerButtonHoldTime);
        digitalWrite(powerButton, LOW);
        delay(powerButtonWaitTime);
        digitalWrite(powerButton, HIGH); // short press power to boil the kettle
        delay(powerButtonWaitTime);
        digitalWrite(powerButton, LOW);
      }
    }
  }
  else {
    Serial.print("No match for message: ");
    Serial.println(inMessage);
  }
  
}

void reconnect() {
  // Loop until we're reconnected to MQTT
  while (!client.connected()) {
    Serial.print("Attempting MQTT connection to broker: ");
    Serial.println(mqtt_server);
    // Attempt to connect
      if (client.connect("Kettle", mqtt_user, mqtt_pass, mqtt_willTopic, 0, true, mqtt_willMessage)) {
        Serial.println("MQTT connected");
        client.publish(mqtt_willTopic, "Online", true);
        client.publish(mqtt_state, mqtt_offMessage); // inital state is off until polling starts
        // Subscribe to our selected topic
        client.subscribe(mqtt_topic);

    }
    else {
      Serial.print("MQTT connection failed, rc=");
      Serial.print(client.state());
      Serial.println(".  Trying again in 5 seconds.");
      // Wait 5 seconds before retrying
      delay(5000);
    }
  }
}

void setup_OTA() {
  // Start up OTA
  
   ArduinoOTA.setPort(8266); // Port defaults to 8266
   ArduinoOTA.setHostname("Kettle"); // Hostname defaults to esp8266-[ChipID]
   ArduinoOTA.setPassword(""); // No authentication by default

  ArduinoOTA.onStart([]() {
    Serial.println("OTA update start");
    client.publish(mqtt_topic, "OTA update start");
  });
  ArduinoOTA.onEnd([]() {
    Serial.println("\nOTA update complete");
    client.publish(mqtt_topic, "OTA update complete");
  });
  ArduinoOTA.onProgress([](unsigned int progress, unsigned int total) {
    Serial.printf("Progress: %u%%\r", (progress / (total / 100)));
  });
  ArduinoOTA.onError([](ota_error_t error) {
    Serial.printf("OTA Error[%u]: ", error);
    if (error == OTA_AUTH_ERROR) Serial.println("Auth Failed");
    else if (error == OTA_BEGIN_ERROR) Serial.println("Begin Failed");
    else if (error == OTA_CONNECT_ERROR) Serial.println("Connect Failed");
    else if (error == OTA_RECEIVE_ERROR) Serial.println("Receive Failed");
    else if (error == OTA_END_ERROR) Serial.println("End Failed");
  });
  ArduinoOTA.begin();
  Serial.println("OTA firmware update ready");  
}

void setup() {
  // Deal with hardware first
  pinMode(powerButton, OUTPUT);
  pinMode(powerMonitor, INPUT_PULLUP);

  Serial.begin(9600);
  Serial.println("");
  Serial.println("Hardware initialized, starting program load");

  // Start up networking
  setup_wifi();
  // Start up MQTT
  long port;
  port = strtol(mqtt_port,NULL,10);
  client.setServer(mqtt_server, port);
  client.setCallback(callback);

  // Start up OTA
  setup_OTA();
}

void loop() {
  if (!client.connected()) {
    reconnect();
  }

  // MQTT client loop
  client.loop();

  // OTA loop
  ArduinoOTA.handle();
  
  if (pulseIn(powerMonitor, LOW, 10000) > 0) { // if there is a pulse on the LED then the Kettle is on
    kettlePreviousState = kettleState;
    kettleState = 1;
  }
  else {  // no pulse then the kettle is off
    kettlePreviousState = kettleState;
    kettleState = 0;
  }
  
  if (kettleState != kettlePreviousState ) {  // only publish the state if it has changed
    if (kettleState == 1) {
      client.publish(mqtt_state, mqtt_onMessage, false);
    }
    else {
      client.publish(mqtt_state, mqtt_offMessage, false);
    }
  }
  delay(490);
}






